@@ -27,7 +27,7 @@ Config:
         Leading characters to include in nicks of puppet users.
 """
 
-from asyncio import (CancelledError, Future, ensure_future,
+from asyncio import (CancelledError, Event, Future, ensure_future,
                      open_connection, sleep, TimeoutError, wait_for)
 import logging
 import re
@@ -186,9 +186,9 @@ class IRCRichText(immp.RichText):
 class IRCUser(immp.User):
 
     @classmethod
-    def from_id(cls, irc, id_):
+    def from_id(cls, irc, id_, real_name=None):
         nick = id_.split("!", 1)[0]
-        return immp.User(id_=id_, plug=irc, username=nick, raw=id_)
+        return immp.User(id_=id_, plug=irc, username=nick, real_name=real_name, raw=id_)
 
     @classmethod
     def from_who(cls, irc, line):
@@ -331,7 +331,7 @@ class IRCClient:
     _nick_bad_chars = re.compile(r"[^A-Z0-9^_\-\[\]{}\|]", re.I)
 
     def __init__(self, plug, host, port, ssl, nick, password=None,
-                 user=None, name=None, listener=False):
+                 user=None, name=None, on_connect=None, on_receive=None):
         self._plug = plug
         # Server parameters for (re)connections.
         self._host = host
@@ -341,11 +341,16 @@ class IRCClient:
         self._password = password
         self._user = user
         self._name = name
-        self._listener = listener
+        self._on_connect = on_connect
+        self._on_receive = on_receive
         # Cache our user-host as seen by the server.
         self._mask = None
         # Connection streams.
-        self._reader = self._writer = self._task = None
+        self._reader = self._writer = self._read_task = None
+        self._closing = False
+        # Background task to send pings if we don't receive any for a while.
+        self._live = Event()
+        self._live_task = None
         # Store channel and nick prefixes for matching and cleanup.
         self.types = ""
         self.prefixes = ""
@@ -385,20 +390,50 @@ class IRCClient:
         else:
             return None
 
+    @property
+    def name(self):
+        return self._name
+
+    @property
+    def closing(self):
+        return self._closing
+
     async def _read_loop(self):
         while True:
-            raw = await self._reader.readline()
+            try:
+                raw = await self._reader.readline()
+            except ConnectionError:
+                log.debug("Client %r disconnected", self._nick, exc_info=True)
+                break
             if not raw:
-                # Connection has been closed.
-                self._writer.close()
-                self._reader = self._writer = None
+                log.debug("Client %r reached EOF", self._nick)
                 break
             line = Line.parse(raw.decode().rstrip("\r\n"))
+            if line.command == "QUIT" and line.source == self.nickmask:
+                log.debug("Client %r quitting", self._nick)
+                break
             log.debug("Client %r received line: %r", self._nick, line)
             await self._handle(line)
-        log.debug("Reconnecting in 3 seconds")
-        await sleep(3)
-        await self.connect()
+        self._writer.close()
+        self._reader = self._writer = None
+        if not self._closing:
+            ensure_future(self._reconnect("Disconnected"))
+
+    async def _keepalive_loop(self):
+        while True:
+            try:
+                await wait_for(self._live.wait(), timeout=120)
+            except TimeoutError:
+                # Nothing from the server in 2 minutes, send a PING.
+                try:
+                    await self._wait(Line("PING", self._nick), success=("PONG",))
+                except TimeoutError:
+                    log.debug("Client %r ping timeout", self._nick)
+                    break
+            finally:
+                self._live.clear()
+        if not self._closing:
+            ensure_future(self._reconnect("Disconnected"))
 
     def _write(self, *lines):
         for line in lines:
@@ -407,20 +442,21 @@ class IRCClient:
 
     async def _wait(self, *lines, success=(), fail=(), collect=()):
         wait = Wait(success, fail, collect)
-        log.debug("Adding wait: %r", wait)
+        log.debug("Client %r adding wait: %r", self._nick, wait)
         self._waits.append(wait)
         self._write(*lines)
         try:
             result = await wait_for(wait, 10)
-            log.debug("Completing wait: %r", wait)
+            log.debug("Client %r completing wait: %r", self._nick, wait)
         except TimeoutError:
-            log.warning("Timeout on wait: %r", wait, exc_info=True)
+            log.warning("Client %r timed out on wait: %r", self._nick, wait)
             raise
         finally:
             self._waits.remove(wait)
         return result
 
     async def _handle(self, line):
+        self._live.set()
         # Route lines to any waits listening for them.
         for wait in self._waits:
             if wait.wants(line):
@@ -505,12 +541,13 @@ class IRCClient:
                     log.debug("Replacing %s with %s in %s members", old, new, name)
                     members.remove(old)
                     members.add(new)
-        if self._listener:
-            await self._plug._handle(line)
+        if self._on_receive:
+            await self._on_receive(line)
 
     async def connect(self):
+        self._closing = False
         self._reader, self._writer = await open_connection(self._host, self._port, ssl=self._ssl)
-        self._task = ensure_future(self._read_loop())
+        self._read_task = ensure_future(self._read_loop())
         if self._password:
             self._write(Line("PASS", self._password))
         await self._wait(Line("USER", self._user, "0", "*", self._name),
@@ -519,20 +556,40 @@ class IRCClient:
         for user in await self.who(self._nick):
             if user.username == self._nick:
                 self._mask = user.id.split("!", 1)[-1]
+        self._live_task = ensure_future(self._keepalive_loop())
+        if self._on_connect:
+            await self._on_connect()
 
     async def disconnect(self, msg):
+        self._closing = True
         for wait in self._waits:
             wait.cancel()
         self._waits.clear()
-        if self._task:
-            self._task.cancel()
-            self._task = None
-        self._reader = None
+        if self._live_task:
+            self._live_task.cancel()
+            self._live_task = None
         if self._writer:
             self._write(Line("QUIT", msg))
-            await self._writer.drain()
+            if self._read_task:
+                try:
+                    await wait_for(self._read_task, 10)
+                except TimeoutError:
+                    log.debug("Server didn't hangup after QUIT")
+        if self._read_task:
+            self._read_task.cancel()
+            self._read_task = None
+        if self._writer:
             self._writer.close()
             self._writer = None
+        self._reader = None
+        self.users.clear()
+        self.members.clear()
+
+    async def _reconnect(self, msg):
+        await self.disconnect(msg)
+        log.debug("Reconnecting in 3 seconds")
+        await sleep(3)
+        await self.connect()
 
     async def who(self, name):
         if name in self.users:
@@ -626,12 +683,9 @@ class IRCPlug(immp.Plug):
                                  self.config["server"]["password"],
                                  "immp",
                                  self.config["user"]["real-name"],
-                                 True)
+                                 self._connected,
+                                 self._handle)
         await self._client.connect()
-        for channel in self.host.channels.values():
-            if channel.plug == self and channel.source.startswith("#"):
-                self._joins.add(channel.source)
-                await self._client.join(channel.source)
 
     async def stop(self):
         await super().stop()
@@ -654,10 +708,14 @@ class IRCPlug(immp.Plug):
         user = self.get_user(username)
         if user:
             return user
+        for client in self._puppets.values():
+            if username == client.nick:
+                return IRCUser.from_id(self, client.nickmask, client.name)
         for user in await self._client.who(username):
             if user.username == username:
                 return user
-        return None
+        else:
+            return None
 
     async def user_is_system(self, user):
         if user.id == self._client.nickmask:
@@ -684,6 +742,8 @@ class IRCPlug(immp.Plug):
         names = set()
         for line in raw:
             names.update(name.lstrip(self._client.prefixes) for name in line.args[3].split())
+        for client in self._puppets.values():
+            names.remove(client.nick)
         return [immp.Channel(self, name) for name in names
                 if name != self.config["user"]["nick"]]
 
@@ -697,6 +757,8 @@ class IRCPlug(immp.Plug):
         return channel.source
 
     async def channel_members(self, channel):
+        if self._client.closing:
+            return None
         try:
             nicks = self._client.members[channel.source]
         except KeyError:
@@ -713,13 +775,22 @@ class IRCPlug(immp.Plug):
     async def channel_remove(self, channel, user):
         self._client.kick(channel.source, user.username)
 
+    async def _connected(self):
+        for channel in self.host.channels.values():
+            if channel.plug == self and channel.source.startswith("#"):
+                self._joins.add(channel.source)
+                await self._client.join(channel.source)
+
     async def _handle(self, line):
         if line.command in ("JOIN", "PART", "KICK", "PRIVMSG"):
             sent = await IRCMessage.from_line(self, line)
+            # Suppress initial joins and final parts.
             if sent.joined and sent.joined[0].id == self._client.nickmask:
                 if sent.channel.source in self._joins:
                     self._joins.remove(sent.channel.source)
                     return
+            elif sent.left and sent.left[0].id == self._client.nickmask and self._client.closing:
+                return
             puppets = [client.nickmask for client in self._puppets.values()]
             # Don't yield messages sent by puppets, or for puppet kicks.
             if sent.user.id in puppets:
